/*
 *   Displays all instances of a given concept.
 */

#include <stdio.h>
#include <dirent.h>
#include "X11/Intrinsic.h"
#include "X11/StringDefs.h" 
#include "X11/Shell.h" 
#include "X11/Xaw/Command.h"
#include "X11/Xaw/Label.h"
#include "X11/Xaw/Box.h"
#include "X11/Xaw/Cardinals.h"
#include "X11/Xaw/Dialog.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <sys/types.h>
#include <math.h>

#define FALSE 0
#define TRUE 1
#define SQRT_NUMSCENES 3   /* Display SQRT_NUMSCENESxSQRT_NUMSCENES scenes */
#define DIM 20         /* Image is DIMxDIM (was 10). */
#define N 100          /* Max number of units allowed in net */
#define X_OFF 0      /* X offset from scene window border */
#define Y_OFF 0      /* Y offset from scene window border */
#define N_OBJS 27      /* Number of objects */


/*  X-related global variables */

Display              *disptr;
Window               win; 
XEvent               event;
GC                   gc;
XFontStruct          *font;
XGCValues            values;
XWindowAttributes    attr;
XSetWindowAttributes set_attr;
XImage               img, *subimg;
XColor               Grey[256];          /* Grey scale */
Pixmap               pmap;
XSetWindowAttributes set_attr;

int   Window_width, Scene_width,   Window_height, Scene_height;
int   CreatedNewFile;
int   ScreenInitialized;

float TrainingResults[N_OBJS];

/*  Object specification */

int ChoseCurObj;
char TRfname[80];
char LMfname[80];
int TRx,TRy,LMx,LMy;
char CurObjfname[80];
XImage CurObjimage;
static XImage TRimage = {
   0, 0,			/* width, height */
   0, XYBitmap, NULL,		/* xoffset, format, data */
   LSBFirst, 8,    		/* byte-order, bitmap-unit */
   LSBFirst, 8, 1		/* bitmap-bit-order, bitmap-pad, depth */
   };
static XImage LMimage = {
   0, 0,			/* width, height */
   0, XYBitmap, NULL,		/* xoffset, format, data */
   LSBFirst, 8,    		/* byte-order, bitmap-unit */
   LSBFirst, 8, 1		/* bitmap-bit-order, bitmap-pad, depth */
   };
static XImage TRNameImage = {
   0, 0,			/* width, height */
   0, XYBitmap, NULL,		/* xoffset, format, data */
   LSBFirst, 8,    		/* byte-order, bitmap-unit */
   LSBFirst, 8, 1		/* bitmap-bit-order, bitmap-pad, depth */
   };
static XImage LMNameImage = {
   0, 0,			/* width, height */
   0, XYBitmap, NULL,		/* xoffset, format, data */
   LSBFirst, 8,    		/* byte-order, bitmap-unit */
   LSBFirst, 8, 1		/* bitmap-bit-order, bitmap-pad, depth */
   };

static XrmOptionDescRec options[] = 
{
    {"-hspace",	"*Box.hSpace",	XrmoptionSepArg,	NULL},
    {"-vspace",	"*Box.vSpace",	XrmoptionSepArg,	NULL},
};

char SCfname[80];                /* Name of scene file */

extern void exit();

Pixmap    grid_pmap, p, pmap;

int   ShowResults;               /* Show results of training?  */
int   ShowPositives;
int   JustReadLM;
int   GridDrawn;
int   Window_width;
int   Window_height;

Widget toplevel;

main(argc, argv)
     int argc;
     char **argv;
{
  int ing;                  /* Needed for command line parsing */
  extern char *optarg;      /* Needed for command line parsing */

  int old_argc;
  char **old_argv;

  float f;
  float UnitScreen();
  char inc;
  char tmpfname[80];
  char pos_or_neg[80];
  
  int read_ok, ix;
  int WriteFile;
  int bitmap_size;
  int cmd_line_TR;
  int cmd_line_LM;

  int status, x,y, i,j;

  static XtCallbackProc Quit();
  static XtCallbackProc RepaintScreen();
  XtAppContext app_con;

  DIR *dirp, *opendir();
  FILE *infile, *fopen();
  struct dirent *entryp, *readdir();


  old_argc = argc;
  old_argv = argv;

  ShowResults = FALSE;
  ChoseCurObj = FALSE;
  CreatedNewFile = FALSE;
  ScreenInitialized = FALSE;
  SCfname[0] = '*';
  TRfname[0] = '*';
  LMfname[0] = '*';
  cmd_line_TR = FALSE;
  cmd_line_LM = FALSE;

  /* Read in options */

  while ((ing = getopt(old_argc,old_argv,"c:r:")) != EOF)
    switch (ing)
      {
      case 'c': strcpy(SCfname, optarg);    break;
      case 'r': 
	strcpy(pos_or_neg, optarg);
	ShowPositives = ((pos_or_neg[0] == 'p') || (pos_or_neg[0] == 'P'));
	ShowResults = TRUE;         
	break;
      } 

  if (SCfname[0] == '*')
    {
      printf (">> Error:  No concept specified.  Use -c <concept_name>.\n");
      exit(1);
    }

  if (ShowResults)
    {
      sprintf (tmpfname, "%s", pos_or_neg);
      infile = fopen( tmpfname, "r");

      ix = 3;
      while (ix >= 0)
	{
	  fscanf(infile, "%d %f", &ix, &f);
	  if (ix >= 0)
	    TrainingResults[ix] = f;
	}
      fclose(infile);
    }

  /*
   *  Kind of gross.  Had to have this guy in here just so I could use
   *    the XmuReadBitmapFile routine (see below).   It's not a widget
   *    based program at all.
   */

  toplevel = XtAppInitialize(&app_con, "Xbuttonbox",
			     options, XtNumber(options),
			     &argc, argv, NULL, NULL, ZERO);

  Scene_width = 150;
  Scene_height = 150;
  Window_width = (SQRT_NUMSCENES*Scene_width)+((SQRT_NUMSCENES+1)*5);
  Window_height = (SQRT_NUMSCENES*Scene_height)+((SQRT_NUMSCENES+1)*5);
  disptr=XOpenDisplay(NULL);
  if (!(font = XLoadQueryFont(disptr, "9x15")))
    {
      printf ("Error:  Cannot load font.\n");
      exit(1);
    }
  i = ((ShowPositives) ? 0 : 495);      /* Where do we place window? */
  j = 0;

  /*  Later decided to shift to middle of screen. */

  i += 300;
  j += 300;

  win=XCreateSimpleWindow(disptr,RootWindow(disptr,0), i,j,
			  Window_width, Window_height, 5,0,
			  WhitePixel(disptr,0));
  sprintf(tmpfname,"%s",&SCfname[7]);      /* Strip off "Scenes/" prefix. */
  XStoreName(disptr,win,tmpfname);

  if (ShowResults)
    {
      set_attr.override_redirect = TRUE;
      XChangeWindowAttributes(disptr,win,CWOverrideRedirect,&set_attr); 
    }

  XMapWindow(disptr,win);
  gc = DefaultGC(disptr, 0);
  XSetForeground(disptr,gc,BlackPixel(disptr,0));
  XSetBackground(disptr,gc,WhitePixel(disptr,0));
  status = ReadBitmapFile ("Bitmaps/TRName",&TRNameImage);
  status = ReadBitmapFile ("Bitmaps/LMName",&LMNameImage);

  if (ShowResults)
    {
      read_ok = TRUE;    ix = 0;
      while (read_ok)
	{
	  sprintf(tmpfname,"%s.%d",SCfname,ix);
	  read_ok = ReadScreenFile(tmpfname);
	  if (read_ok)
	    PaintScreen(ix);
	  else if (ix == 0)    /* No such files */
	    {
	      printf ("No such concept.\n");
	      exit(1);
	    }
	  ix++;
	}
      XFlush(disptr);
    }

  XSelectInput(disptr, win, KeyPressMask | ExposureMask );

  inc = 'c';
  while ((inc != 'q') && (inc != 'Q') && (inc != ' '))
    {
      XNextEvent(disptr, &event);
      if (event.type == KeyPress)
	XLookupString(&event, &inc, 1, NULL, NULL);
      else if (event.type == Expose)
	{
	  read_ok = TRUE;    ix = 0;
	  while (read_ok)
	    {
	      sprintf(tmpfname,"%s.%d",SCfname,ix);
	      read_ok = ReadScreenFile(tmpfname);
	      if (read_ok)
		PaintScreen(ix);
	      else if (ix == 0)    /* No such files */
		{
		  printf ("No such concept.\n");
		  exit(1);
		}
	      ix++;
	    }
	  XFlush(disptr);
	}
    }
      
  XCloseDisplay( disptr );
}


/*
 *  Paint the object on the screen.
 */
PaintObject(image,x,y)
     XImage image;
     int x,y;
{
  XSetState (disptr, gc, BlackPixel(disptr,0), WhitePixel(disptr,0), 
	     GXxor, AllPlanes); 
  XPutImage (disptr, win, gc, &image,
	     0, 0, x-(image.width/2), y-(image.height/2), 
	     image.width, image.height);
}


/*
 *  Paint the TR and LM on the screen.
 */
PaintScreen(ix)
     int ix;               /* Scene index */
{
  int x_off, y_off;
  char valstring[10];
  /* float fabs();*/

  /*  We allow SQRT_NUMSCENESxSQRT_NUMSCENES scenes onto a window.  */

  x_off = ((ix % SQRT_NUMSCENES) * (Scene_width + 5)) + 5;
  y_off = ((ix / SQRT_NUMSCENES) * (Scene_height + 5)) + 5;

  /* Paint bounding rectangle, TR and LM in main scene.  */

  XDrawRectangle(disptr,win,gc, x_off,y_off, Scene_width, Scene_width);

  PaintObject(TRimage,TRx+x_off,TRy+y_off); 
  PaintObject(LMimage,LMx+x_off,LMy+y_off);

  /* Figure out where to place the TR, LM labels in the scene.  */

  if (((TRx-LMx)*(TRx-LMx)) > ((TRy-LMy)*(TRy-LMy)))
    {
      if (fabs((float)(TRx-LMx)) > 16.0)
	{
	  PaintObject(TRNameImage,TRx+x_off,y_off+10);
	  PaintObject(LMNameImage,LMx+x_off,y_off+10);
	}
      else
	{
	  if (TRx < LMx)
	    {
	      PaintObject(TRNameImage,TRx+x_off,y_off+10);
	      PaintObject(LMNameImage,TRx+16+x_off,y_off+10);
	    }
	  else
	    {
	      PaintObject(LMNameImage,LMx+x_off,y_off+10);
	      PaintObject(TRNameImage,LMx+16+x_off,y_off+10);
	    }
	}
    }
  
  else
    {
      if (fabs((float)(TRy-LMy)) > 12.0)
	{
	  PaintObject(TRNameImage,x_off+10,TRy+y_off);
	  PaintObject(LMNameImage,x_off+10,LMy+y_off);
	}
      else
	{
	  PaintObject(TRNameImage,x_off+130,TRy+y_off);
	  PaintObject(LMNameImage,x_off+10,LMy+y_off);
	}
    }

  if (ShowResults)
    {
      sprintf(valstring, "%f", TrainingResults[ix]);
      XDrawImageString(disptr,win,gc, 
		       x_off+(Scene_width/3), y_off+(Scene_width-10), 
		       valstring, 8);
    }
}

/*
 *  Read in a screen file.
 */
int ReadScreenFile (fname)
     char *fname;
{
  FILE *infile, *fopen();
  char c;
  int i,j;
  char junk[80];
  int status, success;

  infile = fopen(fname, "r");
  success = (infile != NULL);

  /* If TR and/or LM was specified on cmd line, that overrides this. */

  if (success)
    {

      fscanf (infile, "LM: %d %d %s\n", &LMx, &LMy, LMfname);
      status = ReadBitmapFile (LMfname,&LMimage);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read landmark object file.\n");
	  exit(1);
	}
    


      fscanf (infile, "TR: %d %d %s\n", &TRx, &TRy, TRfname);
      status = ReadBitmapFile (TRfname,&TRimage);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read trajector object file.\n");
	  exit(1);
	}

      fclose(infile);
    }

  return ( success );
}

/*
 *  Read in a bitmap representation of an object.
 */
int ReadBitmapFile (filename,image)
     char *filename;
     XImage *image;
{
    unsigned int width, height;
    int x_hot, y_hot;
    unsigned char *data;
    int status;

    status = XmuReadBitmapDataFromFile (filename, &width, &height, &data,
					&x_hot, &y_hot);
    if (status != BitmapSuccess) return status;

    image->width = width;
    image->height = height;
    image->data = (char *) data;
    image->bytes_per_line = (image->width + 7) / 8;

    return BitmapSuccess;
}


/*
 *  Converts an x- or y- coord to a float representing the same coords
 *    on the unit screen.
 */
float UnitScreen(x)
     int x;
{
  return ( (float) x / (float) Window_width );
}

/*
 *   Given an x- or y-coord (0.0 to 1.0), this rounds that to the nearest
 *     fraction compatible with the DIMxDIM tesselation.
 */
float Nearest(x)
     float x;
{
  /* float rint();*/

  return( rint((float) x * (float) DIM) / (float) DIM );
}

/*  
 *   Displays a pixmap, centered at x,y (0.0 <= x,y <= 1.0).
 */
CenterPixmap(fx,fy, pmap, size)
     float fx,fy;    /* coords to center the pixmap at */
     Pixmap pmap;
     int size;     /* Width, height (assumed equal) of pmap */
{
  int x,y;

  x = (int) (fx * (float) Window_width);
  y = (int) (fy * (float) Window_height);

  XCopyArea(disptr, pmap, win, gc, 0,0, size,size, x-(size/2), y-(size/2));
}


